from ..value_set import ValueSet


class FrailtyDevice(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of devices for durable medical equipment (DME) used by frail patients.

    **Data Element Scope:** This value set may use a model element related to Device.

    **Inclusion Criteria:** Includes concepts that represent  durable medical equipment (DME) or devices for frailty.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS134v10, CMS165v10, CMS131v10, CMS122v10, CMS125v10, CMS130v10
    """

    VALUE_SET_NAME = 'Frailty Device'
    OID = '2.16.840.1.113883.3.464.1003.118.12.1300'
    DEFINITION_VERSION = '20200310'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SNOMEDCT = {
        '23366006',  # Motorized wheelchair device (physical object)
        '23562009',  # Household ventilator, device (physical object)
        '37874008',  # Continuing positive airway pressure unit, device (physical object)
        '58938008',  # Wheelchair device (physical object)
        '66435007',  # Electric bed, device (physical object)
        '71545009',  # Household humidifier, device (physical object)
        '87405001',  # Cane, device (physical object)
        '183240000',  # Self-propelled wheelchair (physical object)
        '183241001',  # Pedal powered wheelchair (physical object)
        '183248007',  # Attendant powered wheelchair (physical object)
        '228869008',  # Manual wheelchair (physical object)
        '261323006',  # Portable oxygen cylinder (physical object)
        '262177002',  # Static oxygen cylinder (physical object)
        '360006004',  # Walking stick (physical object)
        '360008003',  # Commode (physical object)
        '360299009',  # Long cane (physical object)
        '371786002',  # Pressure support ventilator (physical object)
        '391685000',  # Oxygen gas cylinder DD (physical object)
        '391686004',  # Oxygen gas cylinder HD (physical object)
        '391687008',  # Oxygen gas cylinder RD (physical object)
        '391688003',  # Oxygen gas cylinder DF (physical object)
        '391689006',  # Oxygen gas cylinder HX (physical object)
        '391880008',  # Oxygen gas cylinder F (physical object)
        '391881007',  # Oxygen gas cylinder AF (physical object)
        '401953003',  # Oxygen composite cylinder with integral headset 1360 liters (physical object)
        '401954009',  # Oxygen cylinder 1360 liters (physical object)
        '401955005',  # Oxygen cylinder 300 liters (physical object)
        '426160001',  # Oxygen ventilator (physical object)
        '462987000',  # Patient/medical device walker (physical object)
        '463093001',  # Patient/medical device walker, home-use (physical object)
        '464002006',  # Portable ventilator, electric (physical object)
        '464157006',  # Multiple-base walking stick (physical object)
        '464405003',  # Multi-terrain sports wheelchair, attendant/occupant-driven (physical object)
        '464443000',  # Stand-up wheelchair (physical object)
        '464571009',  # Multi-terrain sports wheelchair, electric-motor-driven (physical object)
        '464752005',  # Multi-terrain sports wheelchair, occupant-driven (physical object)
        '465159000',  # Stair-climbing wheelchair (physical object)
        '465556004',  # Single-base walking stick (physical object)
        '465565006',  # Transport wheelchair, collapsible (physical object)
        '465921009',  # Ventilation rocking bed (physical object)
        '466182009',  # Wheelchair, occupant-driven, front-wheels-operated, non-collapsible (physical object)
        '466193006',  # Wheelchair, power-assisted, occupant-controlled, non-collapsible (physical object)
        '466213002',  # Wheelchair, electric-motor-driven, occupant-controlled, manual-steering, collapsible (physical object)
        '466229005',  # Wheelchair, occupant-driven, bimanual-lever-operated, non-collapsible (physical object)
        '466284002',  # Wheelchair, attendant/occupant-driven, bimanual-lever-operated, collapsible (physical object)
        '466316007',  # Wheelchair, combustion-engine-driven, non-collapsible (physical object)
        '466317003',  # Basic walker, foldable (physical object)
        '466322003',  # Wheelchair, power-assisted, attendant/occupant-controlled, non-collapsible (physical object)
        '466331003',  # Wheelchair, attendant/occupant-driven, single-rear-wheel-operated, non-collapsible (physical object)
        '466337004',  # Wheelchair, attendant/occupant-driven, foot-operated, non-collapsible (physical object)
        '466340004',  # Wheelchair, occupant-driven, bimanual-lever-operated, collapsible (physical object)
        '466344008',  # Wheelchair, electric-motor-driven, occupant-controlled, powered-steering, non-collapsible (physical object)
        '466351004',  # Walking chair, foldable (physical object)
        '466364003',  # Wheelchair, attendant/occupant-driven, rear-wheels-operated, collapsible (physical object)
        '466365002',  # Wheelchair, electric-motor-driven, attendant/occupant-controlled, manual-steering, collapsible (physical object)
        '466366001',  # Wheelchair, attendant/occupant-driven, single-lever-operated, non-collapsible (physical object)
        '466378002',  # Wheelchair, electric-motor-driven, occupant-controlled, powered-steering, collapsible (physical object)
        '466381007',  # Wheelchair, attendant-driven, non-collapsible (physical object)
        '466407009',  # Walking stick/seat (physical object)
        '466464004',  # Basic walker, non-foldable (physical object)
        '466466002',  # Wheelchair, occupant-driven, foot-operated, collapsible (physical object)
        '466473007',  # Wheelchair, electric-motor-driven, attendant-controlled, manual-steering, collapsible (physical object)
        '466477008',  # Wheelchair, electric-motor-driven, occupant-controlled, manual-steering, non-collapsible (physical object)
        '466486003',  # Wheelchair, attendant/occupant-driven, bimanual-lever-operated, non-collapsible (physical object)
        '466494005',  # Wheelchair, power-assisted, attendant-controlled, collapsible (physical object)
        '466524001',  # Wheelchair, attendant-driven, collapsible (physical object)
        '466533004',  # Wheelchair, attendant/occupant-driven, foot-operated, collapsible (physical object)
        '466538008',  # Room humidifier (physical object)
        '466550002',  # Wheelchair, attendant/occupant-driven, rear-wheels-operated, non-collapsible (physical object)
        '466553000',  # Wheelchair, occupant-driven, single-lever-operated, collapsible (physical object)
        '466576002',  # Wheelchair, occupant-driven, bimanual-chain-operated, collapsible (physical object)
        '466607002',  # Wheelchair, attendant/occupant-driven, single-lever-operated, collapsible (physical object)
        '466616003',  # Wheelchair, attendant/occupant-driven, single-rear-wheel-operated, collapsible (physical object)
        '466619005',  # Wheelchair, attendant/occupant-driven, bimanual-chain-operated, non-collapsible (physical object)
        '466644002',  # Wheelchair, occupant-driven, bimanual-chain-operated, non-collapsible (physical object)
        '466671002',  # Wheelchair, combustion-engine-driven, collapsible (physical object)
        '466695000',  # Wheelchair, attendant/occupant-driven, single-front-wheel-operated, collapsible (physical object)
        '466699006',  # Wheelchair, power-assisted, attendant/occupant-controlled, collapsible (physical object)
        '466721007',  # Wheelchair, power-assisted, attendant-controlled, non-collapsible (physical object)
        '466739003',  # Wheelchair, occupant-driven, front-wheels-operated, collapsible (physical object)
        '466758007',  # Wheelchair, electric-motor-driven, attendant-controlled, manual-steering, non-collapsible (physical object)
        '466786004',  # Basic electric hospital bed (physical object)
        '466809001',  # Wheelchair, attendant/occupant-driven, bimanual-chain-operated, collapsible (physical object)
        '466813008',  # Wheelchair, electric-motor-driven, attendant-controlled, powered-steering, non-collapsible (physical object)
        '466851008',  # Wheelchair, attendant/occupant-driven, front-wheels-operated, collapsible (physical object)
        '466871004',  # Wheelchair, electric-motor-driven, attendant/occupant-controlled, powered-steering, non-collapsible (physical object)
        '466889003',  # Wheelchair, attendant/occupant-driven, single-front-wheel-operated, non-collapsible (physical object)
        '466926008',  # Wheelchair, occupant-driven, rear-wheels-operated, collapsible (physical object)
        '466927004',  # Wheelchair, electric-motor-driven, attendant/occupant-controlled, powered-steering, collapsible (physical object)
        '466938004',  # Wheelchair, occupant-driven, single-rear-wheel-operated, collapsible (physical object)
        '466947007',  # Wheelchair, occupant-driven, rear-wheels-operated, non-collapsible (physical object)
        '466966007',  # Wheelchair, occupant-driven, foot-operated, non-collapsible (physical object)
        '466986006',  # Walking table (physical object)
        '466989004',  # Wheelchair, attendant/occupant-driven, front-wheels-operated, non-collapsible (physical object)
        '466999009',  # Wheelchair, electric-motor-driven, attendant-controlled, powered-steering, collapsible (physical object)
        '467018005',  # Wheelchair, occupant-driven, single-lever-operated, non-collapsible (physical object)
        '467065004',  # Wheelchair, occupant-driven, single-front-wheel-operated, collapsible (physical object)
        '467077009',  # Wheelchair, electric-motor-driven, attendant/occupant-controlled, manual-steering, non-collapsible (physical object)
        '467095007',  # Wheelchair, power-assisted, occupant-controlled, collapsible (physical object)
        '467137003',  # Wheelchair, occupant-driven, single-rear-wheel-operated, non-collapsible (physical object)
        '467155007',  # Walking chair, non-foldable (physical object)
        '467163008',  # Wheelchair, occupant-driven, single-front-wheel-operated, non-collapsible (physical object)
        '469361005',  # Heat/moisture exchanger, reusable (physical object)
        '469860004',  # All-plastic conventional wheelchair (physical object)
        '470119002',  # Heat/moisture exchanger, single-use (physical object)
        '470174002',  # Heat/moisture exchanger insertable filter (physical object)
        '700593005',  # Heated respiratory humidifier (physical object)
        '700705005',  # Non-heated respiratory humidifier (physical object)
        '700910000',  # Ultrasonic respiratory humidifier (physical object)
        '702172008',  # Home continuous positive airway pressure unit (physical object)
        '702173003',  # Home bilevel positive airway pressure unit (physical object)
        '705404007',  # Walker/Walking frame (physical object)
        '705406009',  # Walker (physical object)
        '705419008',  # Special-function wheelchair (physical object)
        '705421003',  # Sports wheelchair (physical object)
        '705422005',  # Power-driven wheelchair (physical object)
        '705423000',  # Electric-motor-driven wheelchair (physical object)
        '705425007',  # Attendant/occupant-controlled electric-motor-driven wheelchair (physical object)
        '705426008',  # Attendant-controlled electric-motor-driven wheelchair (physical object)
        '705427004',  # Power-assisted wheelchair (physical object)
        '705428009',  # Manual-driven wheelchair (physical object)
        '706180003',  # Respiratory humidifier (physical object)
        '714700001',  # Bilevel positive airway pressure unit hand held (physical object)
        '781471009',  # Gait rehabilitation electronic walker (physical object)
    }


class CardiacPacer(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for a cardiac pacer device.

    **Data Element Scope:** This value set may use a model element related to Device.

    **Inclusion Criteria:** Includes concepts that represent a device that is a cardiac pacer.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS144v10, CMS145v10
    """

    VALUE_SET_NAME = 'Cardiac Pacer'
    OID = '2.16.840.1.113883.3.526.3.1193'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SNOMEDCT = {
        '14106009',  # Cardiac pacemaker, device (physical object)
        '56961003',  # Cardiac transvenous pacemaker, device (physical object)
        '360127006',  # Intravenous cardiac pacemaker system (physical object)
        '360128001',  # Intravenous triggered cardiac pacemaker system (physical object)
        '424921004',  # Permanent cardiac pacemaker, device (physical object)
    }




from typing import List, Tuple

import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms138v6p2 import ClinicalQualityMeasure138v6p2
from canvas_workflow_kit.patient_recordset import (
    BillingLineItemRecordSet,
    InstructionRecordSet,
    InterviewRecordSet,
    MedicationRecordSet,
    ProtocolOverrideRecordSet
)
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestClinicalQualityMeasure138v6p2(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure138v6p2.enabled())

    def test_description(self):
        expected = ('Patients aged 18 years and older, and identified as a tobacco user, '
                    'who have not received tobacco cessation intervention such as counselling, '
                    'referral or medication.')
        self.assertEqual(expected, ClinicalQualityMeasure138v6p2._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS138v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure138v6p2._meta.information)

    def test_change_types(self):
        result = ClinicalQualityMeasure138v6p2._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'billing_line_item',
            'instruction',
            'interview',
            'medication',
            'patient',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_in_initial_population(self):
        start = arrow.get('2018-04-01 13:24:56')
        end = arrow.get('2019-04-01 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient over 18 + patient had a preventive visit
        tests = {
            '2018-03-30': False,
            '2018-04-02': True,
            '2019-02-06': True,
            '2019-03-30': True,
            '2019-04-02': False,
        }
        for on_day, expected in tests.items():
            # context guidance --> always true regardless of any visit
            patient = self.load_patient('cms138v6')
            tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
            self.assertTrue(tested.in_initial_population(), f'{on_day}')

            # context report
            patient = self.load_patient('cms138v6')
            patient.billing_line_items = self.helper_visits([(on_day, '99396')])
            tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
            if expected:
                self.assertTrue(tested.in_initial_population(), f'{on_day}')
            else:
                self.assertFalse(tested.in_initial_population(), f'{on_day}')

        # patient over 18 + patient had two visits
        tests = {
            '2018-03-30': False,
            '2018-04-02': True,
            '2019-02-06': True,
            '2019-03-30': True,
            '2019-04-02': False,
        }
        for on_day, expected in tests.items():
            patient = self.load_patient('cms138v6')
            patient.billing_line_items = self.helper_visits([
                (on_day, '96151'),
                (on_day, '99201'),
            ])
            tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
            if expected:
                self.assertTrue(tested.in_initial_population(), f'{on_day}')
            else:
                self.assertFalse(tested.in_initial_population(), f'{on_day}')

        # patient over 18 + patient had one visit --> false
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2019-02-06', '96151')])
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())

        # patient is not 18 + patient had a preventive visit --> false
        patient = self.load_patient('cms138v6')
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([('2019-02-06', '99396')])
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())

        # patient is not 18 + patient had two visits --> false
        patient = self.load_patient('cms138v6')
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([
            ('2019-02-06', '99396'),
            ('2019-02-06', '99201'),
        ])
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())

    def test_in_initial_population_with_override(self):
        override = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure138v6p2',
            'adjustment': {
                'reference': '2018-10-01T00:00:00Z',
                'cycleDays': 60
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable

        # patient over 18 + patient had a preventive visit
        tests = {
            '2019-01-14': False,
            '2019-01-16': True,
            '2019-03-14': True,
            '2019-04-17': False,
        }
        for on_day, expected in tests.items():
            patient = self.load_patient('cms138v6')
            patient.protocol_overrides = override
            patient.billing_line_items = self.helper_visits([('2019-01-15', '99396')])
            tested = ClinicalQualityMeasure138v6p2(
                patient=patient, now=arrow.get(f'{on_day} 13:24:56'))
            if expected:
                self.assertTrue(tested.in_initial_population(), f'{on_day}')
            else:
                self.assertFalse(tested.in_initial_population(), f'{on_day}')

        # patient over 18 + patient had two visits
        tests = {
            '2019-01-14': False,
            '2019-01-16': True,
            '2019-03-14': True,
            '2019-03-16': False,
        }
        for on_day, expected in tests.items():
            patient = self.load_patient('cms138v6')
            patient.protocol_overrides = override
            patient.billing_line_items = self.helper_visits([
                ('2019-01-15', '96151'),
                ('2019-01-15', '99201'),
            ])
            tested = ClinicalQualityMeasure138v6p2(
                patient=patient, now=arrow.get(f'{on_day} 13:24:56'))
            if expected:
                self.assertTrue(tested.in_initial_population(), f'{on_day}')
            else:
                self.assertFalse(tested.in_initial_population(), f'{on_day}')

        # patient over 18 + patient had one visit --> false
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.billing_line_items = self.helper_visits([('2019-02-06', '96151')])
        tested = ClinicalQualityMeasure138v6p2(
            patient=patient, now=arrow.get('2019-02-26 13:24:56'))
        self.assertFalse(tested.in_initial_population())

        # patient is not 18 + patient had a preventive visit --> false
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([('2019-02-06', '99396')])
        tested = ClinicalQualityMeasure138v6p2(
            patient=patient, now=arrow.get('2019-02-26 13:24:56'))
        self.assertFalse(tested.in_initial_population())

        # patient is not 18 + patient had two visits --> false
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([
            ('2019-02-06', '99396'),
            ('2019-02-06', '99201'),
        ])
        tested = ClinicalQualityMeasure138v6p2(
            patient=patient, now=arrow.get('2019-02-26 13:24:56'))
        self.assertFalse(tested.in_initial_population())

    def test_in_denominator(self):
        start = arrow.get('2018-04-01 13:24:56')
        end = arrow.get('2019-04-01 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient over 18 + patient had a preventive visit + is a user
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # patient over 18 + patient had a preventive visit + is NOT a user
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2018-11-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient over 18 + patient had two visits + is a user
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([
            ('2018-08-29', '96151'),
            ('2018-10-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # patient over 18 + patient had two visits + is NOT a user
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([
            ('2018-08-29', '96151'),
            ('2018-10-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2018-11-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient is not 18 + patient had a preventive visit + is a user --> false
        patient = self.load_patient('cms138v6')
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient is not 18 + patient had two visits --> false
        patient = self.load_patient('cms138v6')
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([
            ('2018-08-29', '96151'),
            ('2018-10-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_denominator_with_override(self):
        override = ProtocolOverrideRecordSet([{
            'protocolKey': 'CMS138v6p2',
            'adjustment': {
                'reference': '2018-10-01T00:00:00Z',
                'cycleDays': 60
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable

        # patient over 18 + patient had a preventive visit + is a user
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.billing_line_items = self.helper_visits([('2019-01-29', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        self.assertTrue(tested.in_denominator())

        # patient over 18 + patient had a preventive visit + is NOT a user
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.billing_line_items = self.helper_visits([('2019-01-29', '99396')])
        patient.interviews = self.helper_tobacco_screening('2018-11-23', '2019-02-05')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        self.assertFalse(tested.in_denominator())

        # patient over 18 + patient had two visits + is a user
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.billing_line_items = self.helper_visits([
            ('2019-01-29', '96151'),
            ('2019-02-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        self.assertTrue(tested.in_denominator())

        # patient over 18 + patient had two visits + is NOT a user
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.billing_line_items = self.helper_visits([
            ('2019-01-29', '96151'),
            ('2019-02-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2018-11-23', '2019-02-05')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        self.assertFalse(tested.in_denominator())

        # patient is not 18 + patient had a preventive visit + is a user --> false
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([('2019-01-29', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        self.assertFalse(tested.in_denominator())

        # patient is not 18 + patient had two visits --> false
        patient = self.load_patient('cms138v6')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([
            ('2019-01-29', '96151'),
            ('2019-02-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        self.assertFalse(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2018-04-01 13:24:56')
        end = arrow.get('2019-04-01 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # population #2
        # - patient over 18 + patient had a preventive visit + is a user + cessation instruction
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        patient.instructions = self.helper_instruction()
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        tested.in_initial_population()
        self.assertTrue(tested.in_numerator())
        # - patient over 18 + patient had a preventive visit + is a user + cessation medication
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        patient.medications = self.helper_medication()
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        tested.in_initial_population()
        self.assertTrue(tested.in_numerator())
        # - patient over 18 + patient had a preventive visit + is a user + No cessation
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, timeframe=timeframe)
        tested.in_initial_population()
        self.assertFalse(tested.in_numerator())

    def test_compute_results(self):
        # patient is under 18
        patient = self.load_patient('cms138v6')
        patient.patient['birthDate'] = '2002-04-01'
        patient.billing_line_items = self.helper_visits([
            ('2019-01-29', '96151'),
            ('2019-02-09', '99201'),
        ])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        patient.instructions = self.helper_instruction()
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(381, result.due_in)

        #  patient over 18 + patient had NO visit
        patient = self.load_patient('cms138v6')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)

        #  patient over 18 + patient had preventive visit + is NOT a user
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2018-11-23', '2019-02-05')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)

        #  patient over 18 + patient had preventive visit + is a user + cessation instruction
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        patient.instructions = self.helper_instruction()
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-17'))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        expected = 'Jenny had a smoking cessation counseling a month ago on 2/6/19.'
        self.assertEqual(expected, ''.join(result.narrative))
        self.assertEqual(326, result.due_in)

        #  patient over 18 + patient had preventive visit + is a user + cessation medication
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        patient.medications = self.helper_medication()
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-30'))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        expected = 'Jenny has been prescribed cessation medication 3 days ago on 3/27/19.'
        self.assertEqual(expected, ''.join(result.narrative))
        self.assertEqual(362, result.due_in)

        #  patient over 18 + patient had NO tobacco screening
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-30'))
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)

        #  patient over 18 + patient had tobacco screening + is user
        patient = self.load_patient('cms138v6')
        patient.billing_line_items = self.helper_visits([('2018-10-09', '99396')])
        patient.interviews = self.helper_tobacco_screening('2019-02-05', '2018-11-23')
        tested = ClinicalQualityMeasure138v6p2(patient=patient, now=arrow.get('2019-03-30'))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        expected = 'Jenny is a current tobacco user, intervention is indicated.'
        self.assertEqual(expected, ''.join(result.narrative))
        self.assertEqual(-1, result.due_in)
        expected = [
            {
                'title': 'Tobacco cessation counseling',
                'narrative': None,
                'command': {
                    'type': 'instruct',
                },
                'rank': 1,
                'key': 'CMS138v6p2_RECOMMEND_CESSATION_COUNSELING',
                'button': 'Plan',
            },
            {
                'title': 'Cessation support medication',
                'narrative': None,
                'command': {
                    'type': 'prescribe',
                },
                'rank': 2,
                'key': 'CMS138v6p2_RECOMMEND_CESSATION_MEDICATION',
                'button': 'Plan',
            },
        ]
        self.helper_compare_recommendations(expected, result.recommendations)

    # helpers
    def helper_compare_recommendations(self, expected, result):
        self.assertEqual(len(expected), len(result))
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])

    def helper_instruction(self) -> InstructionRecordSet:
        return InstructionRecordSet([{
            'noteTimestamp': '2019-02-06 11:54:00+00:00',
            'coding': [{
                'system': 'http://snomed.info/sct',
                'code': '702388001',
            }]
        }])

    def helper_tobacco_screening(self, user_date, non_user_date) -> InterviewRecordSet:
        return InterviewRecordSet(
            [{
                'noteTimestamp': f'{user_date} 11:54:00+00:00',
                'results': [{
                    'codeSystem': 'http://loinc.org',
                    'code': '68535-4'
                }],
                'questionnaires': [{
                    'id': 30,
                    'code': '62541-8',
                    'codeSystem': 'http://loinc.org'
                }],
                'questions': [{
                    'questionResponseId': 5,
                    'code': '711013002',
                    'codeSystem': 'http://snomed.info/sct',
                    'questionnaireId': 30
                }],
                'responses': [{
                    'questionResponseId': 5,
                    'value': 'Current every day smoker',
                    'code': '449868002',
                    'codeSystem': 'http://snomed.info/sct'
                }]
            }, {
                'noteTimestamp': f'{non_user_date} 11:54:00+00:00',
                'results': [{
                    'codeSystem': 'http://loinc.org',
                    'code': '68535-4'
                }],
                'questionnaires': [{
                    'id': 30,
                    'code': '62541-8',
                    'codeSystem': 'http://loinc.org'
                }],
                'questions': [{
                    'questionResponseId': 5,
                    'code': '711013002',
                    'codeSystem': 'http://snomed.info/sct',
                    'questionnaireId': 30
                }],
                'responses': [{
                    'questionResponseId': 5,
                    'value': 'Current every day smoker',
                    'code': '105539002',
                    'codeSystem': 'http://snomed.info/sct'
                }]
            }]
        )  # yapf: disable

    def helper_medication(self) -> MedicationRecordSet:
        return MedicationRecordSet([
            {
                'coding': [
                    {
                        'system': 'http://www.nlm.nih.gov/research/umls/rxnorm',
                        'code': '993550',
                    },
                    {
                        'system': 'http://www.fdbhealth.com/',
                        'code': '558662',
                    },
                ],
                'periods': [{
                    'from': '2019-03-27',
                    'to': None,
                }],
            },
        ])

    def helper_visits(self, visits: List[Tuple[str, str]]) -> BillingLineItemRecordSet:
        return BillingLineItemRecordSet([{
            'created': f'{on_date} 11:54:00+00:00',
            'datetimeOfService': f'{on_date} 11:54:00+00:00',
            'cpt': code,
        } for on_date, code in visits])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['canvas_workflow_kit',
 'canvas_workflow_kit.builtin_cqms',
 'canvas_workflow_kit.builtin_cqms.future',
 'canvas_workflow_kit.builtin_cqms.stubs',
 'canvas_workflow_kit.builtin_cqms.tests',
 'canvas_workflow_kit.builtin_cqms.tests.mixins',
 'canvas_workflow_kit.internal',
 'canvas_workflow_kit.tests',
 'canvas_workflow_kit.tests.get_protocols_unittest',
 'canvas_workflow_kit.tests.sdk',
 'canvas_workflow_kit.tests.value_set',
 'canvas_workflow_kit.value_set',
 'canvas_workflow_kit.value_set.customer_value_sets',
 'canvas_workflow_kit.value_set.v2021',
 'canvas_workflow_kit.value_set.v2022']

package_data = \
{'': ['*'],
 'canvas_workflow_kit.builtin_cqms.stubs': ['.mypy_cache/3.6/*',
                                            '.mypy_cache/3.6/collections/*'],
 'canvas_workflow_kit.builtin_cqms.tests': ['mock_data/cms122v6_diabetes_no/*',
                                            'mock_data/cms122v6_diabetes_yesnotest/*',
                                            'mock_data/cms122v6_diabetes_yeswithtest/*',
                                            'mock_data/cms123v6_diabetes_no/*',
                                            'mock_data/cms123v6_diabetes_yes/*',
                                            'mock_data/cms123v6_diabetes_yesdone/*',
                                            'mock_data/cms125v6_woman_due/*',
                                            'mock_data/cms125v6_woman_mammography/*',
                                            'mock_data/cms125v6_woman_mastectomy/*',
                                            'mock_data/cms130v6_patient_due/*',
                                            'mock_data/cms130v6_patient_satisfied/*',
                                            'mock_data/cms131v6_diabetes_no/*',
                                            'mock_data/cms131v6_diabetes_yesnoreferralreport/*',
                                            'mock_data/cms131v6_diabetes_yeswithreferralreport/*',
                                            'mock_data/cms134v6_diabetes_no/*',
                                            'mock_data/cms134v6_diabetes_yes/*',
                                            'mock_data/cms134v6_diabetes_yes_only/*',
                                            'mock_data/cms138v6/*',
                                            'mock_data/diabetes_mixin/*',
                                            'mock_data/diabetic1/*',
                                            'mock_data/feet1/*',
                                            'mock_data/hcc001v1/*',
                                            'mock_data/hcc002v2_diagnosed/*',
                                            'mock_data/hcc003v1_diagnosed/*',
                                            'mock_data/hcc004v1/*',
                                            'mock_data/hcc005v1/*',
                                            'mock_data/patient/*'],
 'canvas_workflow_kit.tests': ['mock_data/empty.json',
                               'mock_data/empty.json',
                               'mock_data/empty/*',
                               'mock_data/full.json',
                               'mock_data/full.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/billing_line_items.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/conditions.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/imaging_reports.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/immunizations.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/inpatientStay.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/instructions.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/interviews.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/lab_reports.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/medications.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/patient.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/prescriptions.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/procedures.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/protocol_overrides.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referral_reports.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/referrals.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/full/vital_signs.json',
                               'mock_data/partial/*'],
 'canvas_workflow_kit.tests.sdk': ['mock_data/bmi1/*',
                                   'mock_data/bmi1_with_instruction/*',
                                   'mock_data/bmi2/*',
                                   'mock_data/depression1/*',
                                   'mock_data/depression2/*',
                                   'mock_data/diabetic1/*',
                                   'mock_data/diabetic2/*',
                                   'mock_data/diabetic3/*',
                                   'mock_data/diabetic4/*',
                                   'mock_data/example/*',
                                   'mock_data/fall1/*',
                                   'mock_data/feet1/*',
                                   'mock_data/fluvax/*',
                                   'mock_data/highrisk1/*',
                                   'mock_data/nonsmoker/*',
                                   'mock_data/referral/*']}

install_requires = \
['arrow',
 'cached-property>=1.5.0,<1.6.0',
 'click',
 'memoization>=0.4.0,<0.5.0',
 'python-decouple>=3.0.0,<4.0.0',
 'requests>=2.0.0,<3.0.0']

entry_points = \
{'console_scripts': ['canvas-cli = canvas_workflow_kit.canvas_cli:cli']}

setup_kwargs = {
    'name': 'canvas-workflow-kit',
    'version': '0.5.12',
    'description': 'Development kit to empower customization of your Canvas instance',
    'long_description': None,
    'author': 'Canvas Team',
    'author_email': 'engineering@canvasmedical.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AcutePharyngitis, AcuteTonsillitis, AntibioticMedicationsForPharyngitis,
    ComorbidConditionsForRespiratoryConditions, CompetingConditionsForRespiratoryConditions,
    EmergencyDepartmentVisit, EncounterInpatient, Ethnicity, GroupAStreptococcusTest,
    HomeHealthcareServices, HospiceCareAmbulatory, HospitalObservationCareInitial,
    MedicalDisabilityExam, OfficeVisit, OncAdministrativeSex, OutpatientConsultation, Payer,
    PreventiveCareEstablishedOfficeVisit0To17, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesGroupCounseling, PreventiveCareServicesIndividualCounseling,
    PreventiveCareServicesInitialOfficeVisit0To17, PreventiveCareServicesInitialOfficeVisit18AndUp,
    PreventiveCareServicesOther, Race, TelephoneVisits)


class ClinicalQualityMeasure146v9(ClinicalQualityMeasure):
    """
    Appropriate Testing for Pharyngitis

    Description: The percentage of episodes for patients 3 years and older with a diagnosis of
    pharyngitis that resulted in an antibiotic dispensing event and a group A streptococcus (strep)
    test

    Definition: None

    Rationale: Group A streptococcal bacterial infections and other infections that cause
    pharyngitis (which are most often viral) often produce the same signs and symptoms (Shulman et
    al., 2012). The American Academy of Pediatrics, the Centers for Disease Control and Prevention,
    and the Infectious Diseases Society of America all recommend a diagnostic test for Strep A to
    improve diagnostic accuracy and avoid unnecessary antibiotic treatment (Linder et al ., 2005).

    Estimated economic costs of pediatric streptococcal pharyngitis in the United States range from
    $224 million to $539 million per year, including indirect costs related to parental work
    losses. At a higher level, the economic cost of antibiotic resistance varies but has extended
    as high as $20 billion in excess direct healthcare costs, with additional costs to society for
    lost productivity as high as $35 billion a year (2008 dollars) (Pfoh et al, 2008).

    Guidance: This is an episode of care measure that examines all eligible episodes for the
    patient during the measurement period.

    This eCQM is an episode-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms146v9
    """

    title = 'Appropriate Testing for Pharyngitis'

    identifiers = ['CMS146v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Linder, J. A., Bates, D. W., Lee, G. M., et al. (2005). Antibiotic treatment of children with sore throat. JAMA, 294(18), 2315-2322.',
        'Pfoh, E., Wessels, M.R., Goldmann, D., et al. (2008). Burden and economic cost of group A streptococcal pharyngitis. Pediatrics, 121(2), 229-234. doi: 10.1542/peds.2007-0484',
        'Shulman, S. T., Bisno, A. L., Clegg, H. W., et al. (2012). Clinical practice guideline for the diagnosis and management of group A streptococcal pharyngitis: 2012 update by the Infectious Diseases Society of America. Clinical Infectious Diseases, 55(10), E86-E102. doi:10.1093/cid/cis629',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Outpatient, telephone, online assessment, observation, or emergency
        department (ED) visits with a diagnosis of pharyngitis and an antibiotic dispensing event
        among patients 3 years or older
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude episodes where the patient is taking antibiotics in the 30 days prior
        to the episode date.

        Exclude episodes where the patient had a competing comorbid condition during the 12 months
        prior to or on the episode date.

        Exclude episodes when the patient had hospice care overlapping with the measurement period.

        Exclude episodes where the patient had a competing diagnosis within three days after the
        episode date.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: A group A streptococcus test in the seven-day period from three days prior to
        the episode date through three days after the episode date

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Infectious Disease Society of America (2012)

        The Infectious Diseases Society of America "recommends swabbing the throat and testing for
        GAS pharyngitis by rapid antigen detection test (RADT) and/or culture because the clinical
        features alone do not reliably discriminate between GAS and viral pharyngitis except when
        overt viral features like rhinorrhea, cough, oral ulcers, and/or hoarseness are present"
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms125v6_breast_cancer_screening import ClinicalQualityMeasure125v6
from canvas_workflow_kit.patient_recordset import BillingLineItemRecordSet, ProtocolOverrideRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestClinicalQualityMeasure125v6(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_description(self):
        expected = ('Women 50-74 years of age who have not had a mammogram to screen for '
                    'breast cancer within the last 27 months.')
        self.assertEqual(expected, ClinicalQualityMeasure125v6._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS125v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure125v6._meta.information)

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure125v6.enabled())

    def test_version(self):
        self.assertTrue(hasattr(ClinicalQualityMeasure125v6._meta, 'version'))

    def test_change_types(self):
        result = ClinicalQualityMeasure125v6._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'imaging_report',
            'patient',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_had_mastectomy(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is woman between 50-74 --> true
        patient = self.load_patient('cms125v6_woman_due')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.had_mastectomy())

        # patient is woman had a bilateral mastectomy --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.had_mastectomy())

        # patient is woman had a two unilateral mastectomies --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.conditions[0]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.had_mastectomy())

        patient.conditions = patient.conditions + patient.conditions
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.had_mastectomy())

        # patient is woman had a unilateral mastectomy
        # + evidence of a right unilateral mastectomy --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.conditions[0]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.had_mastectomy())

        patient.conditions = patient.conditions + patient.conditions
        patient.conditions[0]['coding'][0]['code'] = '429242008'
        patient.conditions[1]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.had_mastectomy())

        # patient is woman had a unilateral mastectomy
        # + evidence of a left unilateral mastectomy --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.conditions[0]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.had_mastectomy())

        patient.conditions = patient.conditions + patient.conditions
        patient.conditions[0]['coding'][0]['code'] = '429242008'
        patient.conditions[1]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.had_mastectomy())

    def test_first_due_in(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is woman + older than 50 --> none
        patient = self.load_patient('cms125v6_woman_due')
        patient.patient['birthDate'] = '1967-08-22'
        protocol = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(protocol.first_due_in())

        # patient is woman + younger than 50 --> due in
        patient = self.load_patient('cms125v6_woman_due')
        patient.patient['birthDate'] = '1967-08-30'
        protocol = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertEqual(6, protocol.first_due_in())

        # patient is woman + younger than 50 + mastectomy--> none
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.patient['birthDate'] = '1967-08-30'
        protocol = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(protocol.first_due_in())

    def test_first_due_in_with_override(self):
        override = ProtocolOverrideRecordSet([{
            'protocolKey': 'CMS125v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable

        # patient is woman + older than 50 --> none
        patient = self.load_patient('cms125v6_woman_mammography')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '1967-07-22'
        tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
        self.assertIsNone(tested.first_due_in())

        # patient is woman + younger than 50 --> due in
        patient = self.load_patient('cms125v6_woman_mammography')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '1967-08-11'
        tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
        self.assertEqual(6, tested.first_due_in())

        # patient is woman + younger than 50 + mastectomy--> none
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.protocol_overrides = override
        patient.patient['birthDate'] = '1967-08-30'
        tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
        self.assertIsNone(tested.first_due_in())

    def test_in_initial_population(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is woman between 50-74 --> true
        patient = self.load_patient('cms125v6_woman_due')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_initial_population())

        patient.billing_line_items = BillingLineItemRecordSet([])
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())
        tested = ClinicalQualityMeasure125v6(
            patient=patient, timeframe=timeframe, context='report')
        self.assertFalse(tested.in_initial_population())
        tested = ClinicalQualityMeasure125v6(
            patient=patient, timeframe=timeframe, context='guidance')
        self.assertTrue(tested.in_initial_population())

        # patient is not a woman
        patient = self.load_patient('cms125v6_woman_due')
        for sex in ['M', 'O', 'UNK']:
            patient.patient['sexAtBirth'] = sex
            tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
            self.assertFalse(tested.in_initial_population())
        patient.patient['sexAtBirth'] = 'F'
        self.assertTrue(tested.in_initial_population())

        # patient is younger than 51
        patient.patient['birthDate'] = end.shift(years=-51, days=+1).format('YYYY-MM-DD')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())
        patient.patient['birthDate'] = end.shift(years=-51, days=-1).format('YYYY-MM-DD')
        self.assertTrue(tested.in_initial_population())

        # patient is older than 74
        patient.patient['birthDate'] = end.shift(years=-74, days=-1).format('YYYY-MM-DD')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())
        patient.patient['birthDate'] = end.shift(years=-74, days=+1).format('YYYY-MM-DD')
        self.assertTrue(tested.in_initial_population())

    def test_in_initial_population_with_override(self):
        override = ProtocolOverrideRecordSet([{
            'protocolKey': 'CMS125v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable

        # patient is woman between 50-74 --> true
        patient = self.load_patient('cms125v6_woman_due')
        patient.protocol_overrides = override
        tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
        self.assertTrue(tested.in_initial_population())

        # patient is not a woman
        for sex in ['M', 'O', 'UNK']:
            patient.patient['sexAtBirth'] = sex
            tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
            self.assertFalse(tested.in_initial_population())
        patient.patient['sexAtBirth'] = 'F'
        self.assertTrue(tested.in_initial_population())

        # patient is younger than 51
        patient.patient['birthDate'] = '1967-08-06'
        tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
        self.assertFalse(tested.in_initial_population())
        patient.patient['birthDate'] = '1967-08-04'
        self.assertTrue(tested.in_initial_population())

        # patient is older than 74
        patient.patient['birthDate'] = '1944-08-04'
        tested = ClinicalQualityMeasure125v6(patient=patient, now=arrow.get('2018-08-05'))
        self.assertFalse(tested.in_initial_population())
        patient.patient['birthDate'] = '1944-08-06'
        self.assertTrue(tested.in_initial_population())

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is woman between 50-74 --> true
        patient = self.load_patient('cms125v6_woman_due')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # patient is woman had a bilateral mastectomy --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient is woman had a two unilateral mastectomies --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.conditions[0]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        patient.conditions = patient.conditions + patient.conditions
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient is woman had a unilateral mastectomy
        # + evidence of a right unilateral mastectomy --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.conditions[0]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        patient.conditions = patient.conditions + patient.conditions
        patient.conditions[0]['coding'][0]['code'] = '429242008'
        patient.conditions[1]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient is woman had a unilateral mastectomy
        # + evidence of a left unilateral mastectomy --> false
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.conditions[0]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        patient.conditions = patient.conditions + patient.conditions
        patient.conditions[0]['coding'][0]['code'] = '429242008'
        patient.conditions[1]['coding'][0]['code'] = '172043006'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is woman between 50-74 --> false
        patient = self.load_patient('cms125v6_woman_due')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient is woman between 50-74 + mammography --> true
        patient = self.load_patient('cms125v6_woman_mammography')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient is woman between 50-74 + mammography too old --> true
        patient = self.load_patient('cms125v6_woman_mammography')
        patient.imaging_reports[0]['originalDate'] = start.shift(months=-15).format('YYYY-MM-DD')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        patient.imaging_reports[0]['originalDate'] = start.shift(
            months=-15, days=+1).format('YYYY-MM-DD')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

    def test_in_numerator_with_override(self):
        patient = self.load_patient('cms125v6_woman_mammography')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure125v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.imaging_reports[0]['originalDate'] = '2018-08-05'

        tested = ClinicalQualityMeasure125v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=0))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure125v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure125v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=180))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure125v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        self.assertFalse(tested.in_numerator())

    def test_compute_results(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is woman between 50-74 --> due
        patient = self.load_patient('cms125v6_woman_due')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(
            result.narrative,
            'No relevant exams found.\nCurrent screening interval 2 years, 3 months.')
        expected = [
            {
                'title': 'Discuss breast cancer screening and order imaging as appropriate',
                'command': {
                    'type': 'instruct',
                    'filter': {
                        'coding': [
                            {
                                'system': 'hcpcs',
                                'code': ['G0206', 'G0202', 'G0204', ]
                            },
                            {
                                'system': 'loinc',
                                'code': [
                                    '37539-4', '37774-7', '24610-8', '37029-6', '37006-4',
                                    '26351-7', '36642-7',
                                    '26175-0', '37553-5', '39153-2', '46337-2', '26346-7',
                                    '46380-2', '48475-8',
                                    '24605-8', '42416-8', '24606-6', '46335-6', '42415-0',
                                    '26176-8', '48492-3',
                                    '38820-7', '38854-6', '26347-5', '69251-7', '46356-2',
                                    '26348-3', '36962-9',
                                    '46350-5', '37770-5', '37038-7', '39150-8', '46354-7',
                                    '38090-7', '69259-0',
                                    '26291-5', '38072-5', '26177-6', '38070-9', '26287-3',
                                    '24604-1', '46355-4',
                                    '42169-3', '37543-6', '36319-2', '26349-1', '46351-3',
                                    '46338-0', '36626-0',
                                    '37037-9', '37052-8', '39154-0', '37551-9', '37005-6',
                                    '38071-7', '26350-9',
                                    '37775-4', '26289-9', '37769-7', '42168-5', '42174-3',
                                    '46339-8', '69150-1',
                                    '38067-5', '38855-3', '38807-4', '37053-6', '37028-8',
                                    '37016-3', '37030-4',
                                    '37017-1', '37542-8', '37771-3', '37552-7', '37768-9',
                                    '39152-4', '46342-2',
                                    '38091-5', '36627-8', '37773-9', '37772-1', '46336-4',
                                    '37554-3', '36625-2',
                                ],
                            },
                        ],
                    },
                },
                'key': 'CMS125v6_RECOMMEND_MAMMOGRAPHY',
                'rank': 1,
                'button': 'Plan',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient is woman between 50-74 + mammography --> satisfied
        patient = self.load_patient('cms125v6_woman_mammography')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = r'^Jane had a mammography [\w\s]+ on 7/30/18.$'
        self.assertRegex(result.narrative, text)
        self.assertEqual([], result.recommendations)
        self.assertEqual(797, result.due_in)

        # patient is woman had a two unilateral mastectomies --> N/A
        patient = self.load_patient('cms125v6_woman_mastectomy')
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

        # patient is woman under 50 without mastectomies -> N/A + due in
        patient = self.load_patient('cms125v6_woman_due')
        patient.patient['birthDate'] = '1967-12-23'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(121, result.due_in)

        # patient is woman under 50 with mastectomies -> N/A
        patient = self.load_patient('cms125v6_woman_mastectomy')
        patient.patient['birthDate'] = '1967-12-23'
        tested = ClinicalQualityMeasure125v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)

    def test_compute_results_with_override(self):
        patient = self.load_patient('cms125v6_woman_mammography')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure125v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.imaging_reports[0]['originalDate'] = '2018-08-05'

        # mammography done within the last 180 days
        tested = ClinicalQualityMeasure125v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=176))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        self.assertEqual('Jane had a mammography 5 months ago on 8/5/18.', result.narrative)
        self.assertEqual(0, len(result.recommendations))
        self.assertEqual(4, result.due_in)

        # mammography done 181 days after the 2018-02-01
        patient.billing_line_items[0]['created'] = '2018-08-10'  # visit within the 180-day period
        patient.billing_line_items[0]['datetimeOfService'] = '2018-08-10'  # visit within the 180-day period
        tested = ClinicalQualityMeasure125v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(result.narrative,
                         'No relevant exams found.\nCurrent screening interval 6 months.')
        self.assertEqual(1, len(result.recommendations))
        self.assertEqual(-1, result.due_in)

    def helper_exam_recommended(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])
            exp_item = exp['command']['filter']['coding']
            rec_item = recommendation.command['filter']['coding']
            self.assertEqual(len(exp_item), len(rec_item))
            for exp_coding, rec_coding in zip(exp_item, rec_item):
                self.assertEqual(exp_coding['system'], rec_coding['system'])
                self.assertEqual(sorted(exp_coding['code']), sorted(rec_coding['code']))



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AcuteInpatient, AnnualWellnessVisit, CareServicesInLongTermResidentialFacility,
    ChronicKidneyDiseaseStage5, DementiaMedications, Ed, EncounterInpatient, EndStageRenalDisease,
    EsrdMonthlyOutpatientServices, EssentialHypertension, Ethnicity, FrailtyDevice,
    FrailtyDiagnosis, FrailtyEncounter, FrailtySymptom, HomeHealthcareServices,
    HospiceCareAmbulatory, KidneyTransplantRecipient, NonacuteInpatient, NursingFacilityVisit,
    Observation, OfficeVisit, OncAdministrativeSex, Outpatient, Payer, Pregnancy,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race)


class ClinicalQualityMeasure165v9(ClinicalQualityMeasure):
    """
    Controlling High Blood Pressure

    Description: Percentage of patients 18-85 years of age who had a diagnosis of hypertension
    overlapping the measurement period or the year prior to the measurement period, and whose most
    recent blood pressure was adequately controlled (<140/90mmHg) during the measurement period

    Definition: None

    Rationale: High blood pressure (HBP), also known as hypertension, is when the pressure in blood
    vessels is higher than normal (Centers for Disease Control and Prevention [CDC], 2016). The
    causes of hypertension are multiple and multifaceted and can be based on genetic
    predisposition, environmental risk factors, being overweight and obese, sodium intake,
    potassium intake, physical activity, and alcohol use. High Blood Pressure is common, according
    to the National Health and Nutrition Examination Survey (NHANES), approximately 85.7 million
    adults >= 20 years of age had HBP (140/90 mm Hg) between 2011 to 2014 (Crim, 2012). Between
    2011-2014 the prevalence of hypertension (>=140/90 mm Hg) among US adults 60 and older was
    approximately 67.2 percent (Benjamin et al., 2017).

    HBP, known as the “silent killer,” increases risks of heart disease and stroke which are two of
    the leading causes of death in the U.S. (Yoon, Fryar, & Carroll, 2015). A person who has HBP is
    four times more likely to die from a stroke and three times more likely to die from heart
    disease (CDC, 2012) The National Vital Statistics Systems Center for Disease Control and
    Prevention reported that in 2014 there were approximately 73,300 deaths directly due to HBP and
    410,624 deaths with any mention of HBP (CDC, 20145).Between 2004 and 2014 the number of deaths
    due to HBP rose by 34.1 percent (Benjamin et al., 2017). Managing and treating HBP would reduce
    cardiovascular disease mortality for males and females by 30.4 percent and 38.0 percent,
    respectively (Patel et al., 2015).

    The estimated annual average direct and indirect cost of HBP from 2012 to 2013 was $51.2
    billion (Benjamin et al., 2017). Total direct costs of HBP is projected to increase to $200
    billion by 2030 (Benjamin et al., 2017). A study on cost-effectiveness on treating hypertension
    found that controlling HBP in patients with cardiovascular disease and systolic blood pressures
    of >=160 mm Hg could be effective and cost-saving (Moran et al., 2015).

    Many studies have shown that controlling high blood pressure reduces cardiovascular events and
    mortality. The Systolic Blood Pressure Intervention Trial (SPRINT) investigated the impact of
    obtaining a SBP goal of <120 mm Hg compared to a SBP goal of <140 mm Hg among patients 50 and
    older with established cardiovascular disease and found that the patients with the former goal
    had reduced cardiovascular events and mortality (SPRINT Research Group et al., 2015).

    Controlling HBP will significantly reduce the risks of cardiovascular disease mortality and
    lead to better health outcomes like reduction of heart attacks, stroke, and kidney disease
    (James et al., 2014). Thus, the relationship between the measure (control of hypertension) and
    the long-term clinical outcomes listed is well established.

    Guidance: In reference to the numerator element, only blood pressure readings performed by a
    clinician or a remote monitoring device are acceptable for numerator compliance with this
    measure.

    Do not include BP readings:
    -Taken during an acute inpatient stay or an ED visit
    -Taken on the same day as a diagnostic test or diagnostic or therapeutic procedure that
    requires a change in diet or change in medication on or one day before the day of the test or
    procedure, with the exception of fasting blood tests.
    -Reported by or taken by the member

    If no blood pressure is recorded during the measurement period, the patient's blood pressure is
    assumed "not controlled."

    If there are multiple blood pressure readings on the same day, use the lowest systolic and the
    lowest diastolic reading as the most recent blood pressure reading.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms165v9
    """

    title = 'Controlling High Blood Pressure'

    identifiers = ['CMS165v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Diabetes Association. (2019). 10. Cardiovascular disease and risk management: Standards of medical care in diabetes—2019. Diabetes Care, 42(Suppl. 1), S103-S123. https://doi.org/10.2337/dc19S010',
        'Benjamin, E. J., Blaha, M. J., Chiuve, S. E., et al. (2017). Heart disease and stroke statistics-2017 update: A report from the American Heart Association. Circulation, 135(10), e146-e603.doi: 10.1161/CIR.0000000000000485',
        'Centers for Disease Control and Prevention. (2012). Vital signs: Getting blood pressure under control. Retrieved from https://www.cdc.gov/vitalsigns/hypertension/index.html',
        'Centers for Disease Control and Prevention. Division for Heart Disease and Stroke Prevention. (2016). High blood pressure fact sheet. Retrieved from https://www.cdc.gov/dhdsp/data_statistics/fact_sheets/fs_bloodpressure.htm',
        'Centers for Disease Control and Prevention, National Center for Health Statistics. (2015). Underlying Cause of Death 1999-2013 on CDC WONDER Online Database. Data are from the Multiple Cause of Death Files, 1999-2013, as compiled from data provided by the 57 vital statistics jurisdictions through the Vital Statistics Cooperative Program. Retrieved from http://wonder.cdc.gov/ucd-icd10.html',
        'Crim, M. T., Yoon, S. S., Ortiz, E., et al. (2012). National surveillance definitions for hypertension prevalence and control among adults. Circulation: Cardiovascular Quality and Outcomes. 2012, ;5(3), :343-–351. doi: 10.1161/ CIRCOUTCOMES.111.963439.',
        'Moran, A. E., Odden, M. C., Thanataveerat, A., et al.Tzong KY, Rasmussen PW, Guzman D, Williams L, Bibbins-Domingo K, Coxson PG, Goldman L. (2015). Cost-effectiveness of hypertension therapy according to 2014 guidelines. [published correction appears in N Engl J. Med. 2015;372:1677]. New England Journal of Medicine. 2015 ;372, 447-455. doi: 10.1056/NEJMsa1406751. [published correction appears on page 1677]',
        'Patel, S. A., Winkel, M., Ali, M. K., et al. (2015). Cardiovascular mortality associated with 5 leading risk factors: National and state preventable fractions estimated from survey data. Annals of Internal Medicine, 163(4), 245-253. doi: 10.7326/M14-1753',
        'Qaseem, A., Wilt, T. J., Rich, R., et al. (2017). Pharmacologic treatment of hypertension in adults aged 60 years or older to higher versus lower blood pressure targets: A clinical practice guideline from the American College of Physicians and the American Academy of Family Physicians. Annals of Internal Medicine, 166(6), 430-437. Retrieved from https://annals.org/aim/fullarticle/2598413/pharmacologic-treatment-hypertension-adults-aged-60-years-older-higher-versus',
        'SPRINT Research Group, Wright, J. T., Jr., Williamson, J. D., et al. (2015). A randomized trial of intensive versus standard blood-pressure control. New England Journal of Medicine, 373(22), 2103–2116.',
        'U.S. Preventive Services Task Force. (2015). Screening for high blood pressure in adults: U.S. Preventive Services Task Force recommendation statement. Annals of Internal Medicine, 163(10), 778-787. Retrieved from https://annals.org/aim/fullarticle/2456129/screening-high-blood-pressure-adults-u-s-preventive-services-task',
        'Whelton, P. K., Carey, R. M., Aronow, W. S., et al. (2017). Guideline for the prevention, detection, evaluation, and management of high blood pressure in adults: A report of the American College of Cardiology/American Heart Association Task Force on Clinical Practice Guidelines. Journal of the American College of Cardiology. https://doi.org/10.1161/HYP.0000000000000065',
        'Yoon, S. S., Fryar, C. D., & Carroll, M. D. (2015). Hypertension prevalence and control among adults: United States, 2011-2014. NCHS Data Brief No. 220. Hyattsville, MD: National Center for Health Statistics.',
        'Farley TA, Dalal MA, Mostashari F, Frieden TR. Deaths preventable in the US by improvements in the use of clinical preventive services. Am J Prev Med. 2010;38:600-9. Retrieved from https://www.ajpmonline.org/article/S0749-3797(10)00207-2/fulltext',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 18-85 years of age who had a visit and diagnosis of essential
        hypertension overlapping the measurement period or the year prior to the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients with evidence of end stage renal disease (ESRD), dialysis or renal
        transplant before or during the measurement period. Also exclude patients with a diagnosis
        of pregnancy during the measurement period.

        Exclude patients whose hospice care overlaps the measurement period.

        Exclude patients 66 and older who are living long term in an institution for more than 90
        consecutive days during the measurement period.

        Exclude patients 66 and older with advanced illness and frailty because it is unlikely that
        patients will benefit from the services being measured.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients whose most recent blood pressure is adequately controlled (systolic
        blood pressure < 140 mmHg and diastolic blood pressure < 90 mmHg) during the measurement
        period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The U.S. Preventive Services Task Force (2015) recommends
        screening for high blood pressure in adults age 18 years and older. This is a grade A
        recommendation.

        American College of Cardiology/American Heart Association (2017)

        -For adults with confirmed hypertension and known CVD or 10-year ASCVD event risk of 10% or
        higher, a blood pressure target of less than 130/80 mmHg is recommended (Level of evidence:
        B-R (for systolic blood pressures), Level of evidence: C-EO (for diastolic blood pressure))

        -For adults with confirmed hypertension, without additional markers of increased CVD risk,
        a blood pressure target of less than 130/80 mmHg may be reasonable (Note: clinical trial
        evidence is strongest for a target blood pressure of 140/90 mmHg in this population.
        However, observational studies suggest that these individuals often have a high lifetime
        risk and would benefit from blood pressure control earlier in life) (Level of evidence:
        B-NR (for systolic blood pressure), Level of evidence: C-EO (for diastolic blood pressure))

        American College of Physicians and the American Academy of Family Physicians (2017):

        -Initiate or intensify pharmacologic treatment in some adults aged 60 years or older at
        high cardiovascular risk, based on individualized assessment, to achieve a target systolic
        blood pressure of less than 140 mmHg (Grade: weak recommendation, Quality of evidence: low)

        -Initiate or intensify pharmacologic treatment in adults aged 60 years or older with a
        history of stroke or transient ischemic attack to achieve a target systolic blood pressure
        of less than 140 mmHg to reduce the risk of recurrent stroke (Grade: weak recommendation,
        Quality of evidence: moderate)

        American Diabetes Association (2019):

        -For individuals with diabetes and hypertension at higher cardiovascular risk (existing
        atherosclerotic cardiovascular disease or 10-year atherosclerotic cardiovascular disease
        risk >15%), a blood pressure target of <130/80 mmHg may be appropriate, if it can be safely
        attained (Level of evidence: C)-For individuals with diabetes and hypertension at lower
        risk for cardiovascular disease (10-year atherosclerotic cardiovascular disease risk <15%),
        treat to a blood pressure target of <140/90 mmHg (Level of evidence: A)
        """
        pass



#!/usr/bin/env python

import json
import textwrap
import re
import sys

from glob import iglob

import xmltodict

from colorama import Fore, init as colorama_init

from canvas_workflow_kit.internal.string import snakecase

from canvas_workflow_kit.value_set import v2021  # NOQA pylint: disable=unused-import
from canvas_workflow_kit.value_set.value_set import ValueSet

MAX_WIDTH = 99

colorama_init(autoreset=True)


def get_subclasses(cls):
    subclasses = []

    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(get_subclasses(subclass))

    return subclasses


value_sets = get_subclasses(ValueSet)


def value_set_by_oid(oid):
    for value_set in value_sets:
        if value_set.OID == oid:
            return value_set

    print(Fore.RED + f'No ValueSet found for OID: {oid}')


def wrap_and_indent(string, prefix=None, width=MAX_WIDTH, indent=4):
    if prefix:
        string = f'{prefix}: {string}'

    wrapped = []

    for line in string.splitlines():
        wrapped.append('\n'.join(
            textwrap.wrap(line, break_long_words=False, width=width - indent)))

    return textwrap.indent('\n'.join(wrapped), ' ' * indent)


def format_list(items, indent=4):
    return '\n'.join(
        "{}'{}',".format(' ' * indent, re.sub(r'\r|\n|\s+', ' ', item.replace("'", "\\'")))
        for item in items)


def v(element):
    if not element:
        return None

    if '@value' in element:
        return element['@value']

    if '@code' in element:
        return element['@code']

    if '@root' in element:
        return element['@root']

    if 'item' in element:
        return v(element['item'])

    if 'part' in element:
        return v(element['part'])

    return element


def d(element):
    return json.dumps(element, indent=2)


def ensure_list(element):
    if not isinstance(element, list):
        return [element]

    return element


def generate(directory, year):
    with open('./builtin_cqms/stub_template.py.txt') as stub:
        stub_template = stub.read()

    oids = None
    qmd = None

    for xml_file in iglob(f'./data/cqms/{directory}/**/CMS*.xml'):
        if 'SimpleXML' in xml_file:
            continue

        print(Fore.GREEN + f'Processing "{xml_file}"')

        with open(xml_file) as file:
            contents = file.read()

        qmd = xmltodict.parse(contents)['QualityMeasureDocument']

        authors = []
        cqm_number = None
        cqm_version = v(qmd['versionNumber']).split('.')[0]
        data_criteria = None
        oids = []
        qmd_value_sets = set()
        references = []
        short_description = v(qmd['text'])
        title = v(qmd['title'])

        for author in ensure_list(qmd['author']):
            authors.append(
                v(author['responsibleParty']['representedResponsibleOrganization']['name']))

        for subject_of in qmd['subjectOf']:
            code = subject_of['measureAttribute']['code']
            value = subject_of['measureAttribute']['value']

            if v(code.get('originalText')) == 'eCQM Identifier (Measure Authoring Tool)':
                cqm_number = v(value)

            if v(code) == 'RAT':
                rationale = v(value)

            if v(code) == 'CRS':
                recommendation = v(value)

            if v(code) == 'DEF':
                definition = v(value)

            if v(code) == 'REF':
                references.append(v(value))

            if v(code) == 'GUIDE':
                guidance = v(value)

            if v(code) == 'IPOP':
                initial_population = v(value)

            if v(code) == 'DENOM':
                denominator = v(value)

            if v(code) == 'DENEX':
                denominator_exclusions = v(value)

            if v(code) == 'DENEXCEP':
                denominator_exceptions = v(value)

            if v(code) == 'NUMER':
                numerator = v(value)

            if v(code) == 'NUMEX':
                numerator_exclusions = v(value)

        cms_id = f'{cqm_number}v{cqm_version}'

        # Heart Failure (HF): Angiotensin-Converting Enzyme (ACE) Inhibitor or Angiotensin Receptor
        # Blocker (ARB) Therapy for Left Ventricular Systolic Dysfunction (LVSD)

        snake_title = re.sub(r'\([^()]*?\)', ' ', title)
        snake_title = (snake_title
                       .lower()
                       .replace(' - ', ' ')
                       .replace(':', '')
                       .replace('%', ' percent')
                       .replace('>', 'gt')
                       .replace('<', 'lt')
                       .replace('.', '_')
                       .replace('/', '_')
                       .strip())  # yapf: disable
        snake_title = re.sub(r'\s+', ' ', snake_title)
        snake_title = snakecase(snake_title)

        denominator_body = 'pass'

        if denominator == 'Equals Initial Population':
            denominator_body = 'return self.in_initial_population()'

        for component in ensure_list(qmd['component']):
            if 'dataCriteriaSection' in component:
                data_criteria = component['dataCriteriaSection']

        if not data_criteria:
            print(Fore.RED + 'dataCriteriaSection not found')
            sys.exit(1)

        def add_oid(oid):
            if oid:
                oids.append(oid)

                return True

            return False

        def handle_observation_criteria(observation):
            value = observation.get('value')
            code = observation.get('code')
            participation = observation.get('participation')

            if not value and not code:
                print(Fore.RED + 'Unknown observationCriteria type:', d(observation))
                return

            value_set_added = False

            if value:
                try:
                    add_oid(value['@valueSet'])
                    value_set_added = True
                except KeyError:
                    # BAG: nothing useful if there's no valueSet
                    pass

            if participation:
                try:
                    add_oid(participation['role']['playingEntity']['code']['@valueSet'])
                    value_set_added = True
                except KeyError:
                    if not value_set_added:
                        print(Fore.RED + 'Unable to get observationCriteria participation code', d(observation))

            if code:
                try:
                    add_oid(code['@valueSet'])
                except KeyError:
                    if (
                            not value_set_added and
                            'Patient Characteristic' not in v(observation['title'])
                    ):
                        print(Fore.RED + 'Unable to get observationCriteria code', d(observation))

            handle_participants(observation.get('participation', []))

        def handle_participants(participants):
            if not participants:
                return

            participants = ensure_list(participants)

            found_participant = False

            for participant in participants:
                role = participant['role']

                if 'playingDevice' in role:
                    if add_oid(role['playingDevice']['code'].get('@valueSet')):
                        found_participant = True

                if 'playingEntity' in role:
                    if add_oid(role['playingEntity']['code'].get('@valueSet')):
                        found_participant = True

                if 'playingMaterial' in role:
                    if add_oid(role['playingMaterial']['code'].get('@valueSet')):
                        found_participant = True

                if 'playingManufacturedMaterial' in role:
                    if add_oid(role['playingManufacturedMaterial']['code'].get('@valueSet')):
                        found_participant = True

            if not found_participant:
                print(Fore.YELLOW + 'participant not found', d(participants))

        for entry in ensure_list(data_criteria['entry']):
            if 'actCriteria' in entry:
                handle_participants(entry['actCriteria'].get('participation', []))

                if 'code' in entry['actCriteria']:
                    try:
                        add_oid(entry['actCriteria']['code']['@valueSet'])
                    except KeyError:
                        print(Fore.RED + 'Unable to get actCriteria code', d(entry['actCriteria']))
                elif 'outboundRelationship' in entry['actCriteria']:
                    try:
                        add_oid(entry['actCriteria']['outboundRelationship']['observationCriteria']['value']['@valueSet'])
                    except KeyError:
                        print(Fore.RED + 'Unable to get actCriteria code', d(entry['actCriteria']))
                else:
                    print(Fore.YELLOW + 'actCriteria', d(entry['actCriteria']))
            elif 'encounterCriteria' in entry:
                try:
                    add_oid(entry['encounterCriteria']['code']['@valueSet'])
                except KeyError:
                    print(Fore.RED + 'Unable to get encounterCriteria code', d(entry['encounterCriteria']))

                if 'dischargeDispositionCode' in entry['encounterCriteria']:
                    add_oid(entry['encounterCriteria']['dischargeDispositionCode']['valueSet'])
            elif 'grouperCriteria' in entry:
                continue
            elif 'observationCriteria' in entry:
                handle_observation_criteria(entry['observationCriteria'])
            elif 'procedureCriteria' in entry:
                if entry['procedureCriteria']['code']:
                    try:
                        add_oid(entry['procedureCriteria']['code'].get('valueSet'))
                    except KeyError:
                        print(Fore.RED + 'Unable to get procedureCriteria code', d(entry['procedureCriteria']))
                else:
                    print(Fore.YELLOW + 'procedureCriteria', entry['procedureCriteria'])
            elif 'substanceAdministrationCriteria' in entry:
                handle_participants(entry['substanceAdministrationCriteria']['participation'])
            elif 'supplyCriteria' in entry:
                handle_participants(entry['supplyCriteria']['participation'])
            else:
                print(Fore.RED + 'Unknown criterion type:', entry)

            for oid in oids:
                value_set = value_set_by_oid(oid)

                qmd_value_sets.add(value_set.__name__)

        with open(f'./builtin_cqms/stubs/cms{cms_id}_{snake_title}.py', 'w') as stub:
            stub.write(
                stub_template.format(
                    authors=format_list(authors, indent=8),
                    class_name=cms_id,
                    cms_url=f'https://ecqi.healthit.gov/ecqm/measures/cms{cms_id}',
                    definition=wrap_and_indent(definition, 'Definition'),
                    denominator=wrap_and_indent(denominator, 'Denominator', indent=8),
                    denominator_body=denominator_body,
                    denominator_exceptions=wrap_and_indent(
                        denominator_exceptions, 'Exceptions', indent=8),
                    denominator_exclusions=wrap_and_indent(
                        denominator_exclusions, 'Exclusions', indent=8),
                    funding_source='',
                    guidance=wrap_and_indent(guidance, 'Guidance'),
                    initial_population=wrap_and_indent(
                        initial_population, 'Initial population', indent=8),
                    numerator=wrap_and_indent(numerator, 'Numerator', indent=8),
                    numerator_exclusions=wrap_and_indent(
                        numerator_exclusions, 'Exclusions', indent=8),
                    rationale=wrap_and_indent(rationale, 'Rationale'),
                    recommendation=wrap_and_indent(
                        recommendation, 'Clinical recommendation', indent=8),
                    references=format_list(references, indent=8),
                    short_description=wrap_and_indent(short_description, 'Description'),
                    title=wrap_and_indent(title),
                    plain_title=title,
                    types=['CQM'],
                    value_sets='\n' + wrap_and_indent(', '.join(sorted(qmd_value_sets))),
                    year=year))


if __name__ == '__main__':
    generate('2021', '2021')



from unittest.mock import PropertyMock, patch

import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.ccp004v1_diabetes import Ccp004v1
from canvas_workflow_kit.patient_recordset import ConditionRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestCcp004v1(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(Ccp004v1.enabled())

    def test_description(self):
        expected = 'All patients with Diagnosis Of Diabetes.'
        self.assertEqual(expected, Ccp004v1._meta.description)

    def test_information(self):
        expected = 'https://canvas-medical.zendesk.com/hc/en-us'
        self.assertEqual(expected, Ccp004v1._meta.information)

    def test_change_types(self):
        result = Ccp004v1._meta.compute_on_change_types
        expected = ['condition']
        self.assertEqual(expected, result)

    def test_version(self):
        self.assertTrue(hasattr(Ccp004v1._meta, 'version'))

    def test_date_of_diagnosis(self):
        start = arrow.get('2019-03-30 13:24:56')
        end = arrow.get('2020-03-30 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no condition -> False
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_diagnosis
        self.assertEqual('', result)

        # not diabetes condition -> False
        patient.conditions = ConditionRecordSet([self.helper_condition('active', '0QPD0JZ')])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_diagnosis
        self.assertEqual('', result)

        # diabetes condition --> active -> date
        patient.conditions = ConditionRecordSet([self.helper_condition('2018-08-20', 'E103213')])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_diagnosis
        self.assertEqual('2018-08-20', result)

        # diabetes condition --> resolved -> ''
        patient.conditions = ConditionRecordSet([self.helper_condition('', 'E103213')])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_diagnosis
        self.assertEqual('', result)

        # several diabetes conditions --> the oldest
        patient.conditions = ConditionRecordSet([
            self.helper_condition('2018-08-19', 'E10341'),
            self.helper_condition('2018-08-17', 'E113513'),
            self.helper_condition('2018-08-18', 'O24113'),
        ])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_diagnosis
        self.assertEqual('2018-08-17', result)

    def test_in_initial_population(self):
        start = arrow.get('2019-03-22 13:24:56')
        end = arrow.get('2020-03-22 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # all patients are in the population
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_initial_population()
        self.assertTrue(result)

    def test_in_denominator(self):
        start = arrow.get('2019-03-22 13:24:56')
        end = arrow.get('2020-03-22 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # all patients are in the denominator
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_initial_population()
        self.assertTrue(result)

    def test_in_numerator(self):
        start = arrow.get('2019-03-30 13:24:56')
        end = arrow.get('2020-03-30 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no condition -> False
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertFalse(result)

        # not diabetes condition -> False
        patient.conditions = ConditionRecordSet([self.helper_condition('2018-08-20', '0QPD0JZ')])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertFalse(result)

        # diabetes condition --> True
        patient.conditions = ConditionRecordSet([self.helper_condition('2018-08-20', 'O24113')])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertTrue(result)

        # diabetes condition --> False
        patient.conditions = ConditionRecordSet([self.helper_condition('', 'O24113')])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertFalse(result)

        # several diabetes conditions --> True
        patient.conditions = ConditionRecordSet([
            self.helper_condition('2018-08-19', 'O24019'),
            self.helper_condition('2018-08-17', 'E133551'),
            self.helper_condition('2018-08-18', 'E10620'),
        ])
        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertTrue(result)

    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp004v1_diabetes.Ccp004v1.date_of_diagnosis',
        new_callable=PropertyMock)
    @patch('canvas_workflow_kit.builtin_cqms.ccp004v1_diabetes.Ccp004v1.in_numerator')
    @patch('canvas_workflow_kit.builtin_cqms.ccp004v1_diabetes.Ccp004v1.in_denominator')
    def test_compute_results(self, in_denominator, in_numerator, date_of_diagnosis):
        start = arrow.get('2019-03-30 13:24:56')
        end = arrow.get('2020-03-30 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # not in_denominator (impossible in theory) -> not_applicable
        in_denominator.return_value = False
        in_numerator.return_value = True
        date_of_diagnosis.return_value = '2017-03-14 13:24:55+00:00'

        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)

        # in_denominator + not in_numerator -> satisfied
        in_denominator.return_value = True
        in_numerator.return_value = False
        date_of_diagnosis.return_value = '2017-03-14 13:24:55+00:00'

        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        self.assertEqual(-1, result.due_in)
        narratives = ['Nicolas has not been diagnosed of diabetes.']
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(0, len(result.recommendations))

        # in_denominator + in_numerator -> due + date of diagnosis
        in_denominator.return_value = True
        in_numerator.return_value = True
        date_of_diagnosis.return_value = '2017-03-14 13:24:55+00:00'

        tested = Ccp004v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(0, result.due_in)
        narratives = ['Nicolas has been diagnosed of diabetes on Tue, Mar 14th 2017.']
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(1, len(result.recommendations))
        recommendation = result.recommendations[0]
        self.assertEqual('CCP004v1_RECOMMEND_CONTACT', recommendation.key)
        self.assertEqual('Contact the patient', recommendation.title)

    def helper_condition(self, active_date: str, icd10: str):
        periods = [{'from': '2016-04-18', 'to': '2017-11-21'}]
        status = 'resolved'
        if active_date:
            status = 'active'
            periods.append({'from': active_date, 'to': None})
        return {
            'clinicalStatus': status,
            'coding': [{
                'code': icd10,
                'system': 'ICD-10'
            }],
            'periods': periods,
        }



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AceInhibitorOrArbIngredient, AceInhibitorOrArbOrArni, AllergyToAceInhibitorOrArb,
    CareServicesInLongTermResidentialFacility, DischargeServicesHospitalInpatient,
    EjectionFraction, Ethnicity, HeartFailure, HomeHealthcareServices,
    IntoleranceToAceInhibitorOrArb, LeftVentricularSystolicDysfunction, ModerateOrSevereLvsd,
    NursingFacilityVisit, OfficeVisit, OncAdministrativeSex, OutpatientConsultation,
    PatientProviderInteraction, Payer, Pregnancy, Race, RenalFailureDueToAceInhibitor)


class ClinicalQualityMeasure135v9(ClinicalQualityMeasure):
    """
    Heart Failure (HF): Angiotensin-Converting Enzyme (ACE) Inhibitor or Angiotensin Receptor
    Blocker (ARB) or Angiotensin Receptor-Neprilysin Inhibitor (ARNI) Therapy for Left Ventricular
    Systolic Dysfunction (LVSD)

    Description: Percentage of patients aged 18 years and older with a diagnosis of heart failure
    (HF) with a current or prior left ventricular ejection fraction (LVEF) < 40% who were
    prescribed ACE inhibitor or ARB or ARNI therapy either within a 12-month period when seen in
    the outpatient setting OR at each hospital discharge

    Definition: Prescribed-Outpatient setting: prescription given to the patient for ACE inhibitor
    or ARB or ARNI therapy at one or more visits in the measurement period OR patient already
    taking ACE inhibitor or ARB or ARNI therapy as documented in current medication list.

    Prescribed-Inpatient setting: prescription given to the patient for ACE inhibitor or ARB or
    ARNI therapy at discharge OR ACE inhibitor or ARB or ARNI therapy to be continued after
    discharge as documented in the discharge medication list.

    LVEF < 40% corresponds to qualitative documentation of moderate dysfunction or severe
    dysfunction.

    Rationale: In the absence of contraindications, ACE inhibitors, ARB, or ARNI therapy is
    recommended for all patients with symptoms of heart failure and reduced left ventricular
    systolic function. Recent trial data have shown ARNI to be superior to ACE inhibitor or ARB
    therapy, however an ACE inhibitor or ARB should still be used for patients in which an ARNI is
    contraindicated. Given that ARNI is a newer therapy, uptake has been slow despite updated
    guideline recommendations that support its use. All pharmacologic agents included in this
    measure have been shown to decrease the risk of death and hospitalization for patients with
    heart failure.

    Guidance: This eCQM is to be reported as patient-based or episode-based, depending on the
    clinical setting. To satisfy this measure, it must be reported for all heart failure patients
    at least once during the measurement period if seen in the outpatient setting. If the patient
    has an eligible inpatient discharge during the measurement period, as defined in the measure
    logic, it is expected to be reported at each hospital discharge.

    The requirement of two or more visits used in Population Criteria 1 is to establish that the
    eligible professional or eligible clinician has an existing relationship with the patient.

    A range value should satisfy the logic requirement for 'Ejection Fraction' as long as the
    ranged observation value clearly meets the less than 40% threshold noted in the denominator
    logic. A range that is inclusive of or greater than 40% would not meet the measure requirement.

    Eligible clinicians who have given a prescription for or whose patient is already taking an
    Angiotensin-Converting Enzyme Inhibitor (ACEI) or Angiotensin Receptor Blocker (ARB) would meet
    performance for this measure. Other combination therapies that consist of an ACEI plus
    diuretic, ARB + neprilysin inhibitor (ARNI), ARB plus diuretic, ACEI plus calcium channel
    blocker, ARB plus calcium channel blocker, or ARB plus calcium channel blocker plus diuretic
    would also meet performance for this measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms135v9
    """

    title = 'Heart Failure (HF): Angiotensin-Converting Enzyme (ACE) Inhibitor or Angiotensin Receptor Blocker (ARB) or Angiotensin Receptor-Neprilysin Inhibitor (ARNI) Therapy for Left Ventricular Systolic Dysfunction (LVSD)'

    identifiers = ['CMS135v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Yancy, C. W., Jessup, M., Bozkurt, B., Butler, J., Casey, D. E., Drazner, M. H., … Wilkoff, B. L. (2013). 2013 ACCF/AHA Guideline for the Management of Heart Failure A Report of the American College of Cardiology Foundation/American Heart Association Task Force on Practice Guidelines. Circulation, 128(16), e240–e327. doi: https://doi.org/10.1161/CIR.0b013e31829e8776',
        'Yancy, C. W., Jessup, M., Bozkurt, B., Butler, J., Casey, D. E., Colvin, M. M., … Westlake, C. (2017). 2017 ACC/AHA/HFSA Focused Update of the 2013 ACCF/AHA Guideline for the Management of Heart Failure: A Report of the American College of Cardiology/American Heart Association Task Force on Clinical Practice Guidelines and the Heart Failure Society of America. Circulation,136(6), e137–e161. doi: 10.1161/cir.0000000000000509',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 18 years and older with a diagnosis of heart failure
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population with a current or prior LVEF < 40%

        Exclusions: None

        Exceptions: Documentation of medical reason(s) for not prescribing ACE inhibitor or ARB or
        ARNI therapy (e.g., hypotensive patients who are at immediate risk of cardiogenic shock,
        hospitalized patients who have experienced marked azotemia, allergy, intolerance, other
        medical reasons).

        Documentation of patient reason(s) for not prescribing ACE inhibitor or ARB or ARNI therapy
        (e.g., patient declined, other patient reasons).

        Documentation of system reason(s) for not prescribing ACE inhibitor or ARB or ARNI therapy
        (e.g., other system reasons).
        """
        pass

    def in_numerator(self):
        """
        Numerator: Patients who were prescribed ACE inhibitor or ARB or ARNI therapy either within
        a 12-month period when seen in the outpatient setting OR at each hospital discharge

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The use of ACE inhibitors is beneficial for patients with prior or
        current symptoms of chronic HFrEF to reduce morbidity and mortality (Class I, Level of
        Evidence A) (ACC/AHA/HFSA, 2017).

        Treatment with an ACE inhibitor should be initiated at low doses [see excerpt from
        guideline table below], followed by gradual dose increments if lower doses have been well
        tolerated... Clinicians should attempt to use doses that have been shown to reduce the risk
        of cardiovascular events in clinical trials. If these target doses of an ACE inhibitor
        cannot be used or are poorly tolerated, intermediate doses should be used with the
        expectation that there are likely to be only small differences in efficacy between low and
        high doses. Abrupt withdrawal of treatment with an ACE inhibitor can lead to clinical
        deterioration and should be avoided (ACCF/AHA, 2013).

        Drugs Commonly Used for Stage C HFrEF (abbreviated to align with focus of measure to
        include only ACE inhibitors and ARB therapy)
        Drug                       Initial Daily Dose(s)                           Maximum Dose(s)
        Mean Doses Achieved in
        Clinical Trials
        ACE Inhibitors
           Captopril              6.25 mg 3 times                               50 mg 3 times
        122.7 mg/d
           Enalapril               2.5 mg twice                                    10 to 20 mg
        twice        16.6 mg/d
           Fosinopril             5 to 10 mg once                               40 mg once
        N/A
           Lisinopril              2.5 to 5 mg once                              20 to 40 mg once
        32.5 to 35.0 mg/d
           Perindopril           2 mg once                                        8 to 16 mg once
        N/A
           Quinapril              5 mg twice                                       20 mg twice
        N/A
           Ramipril               1.25 to 2.5 mg once                         10 mg once
        N/A
           Trandolapril          1 mg once                                        4 mg once
        N/A
        Angiotensin Receptor Blockers
           Candesartan         4 to 8 mg once                                32 mg once
        24 mg/d
           Losartan              25 to 50 mg once                             50 to 150 mg once
        129 mg/d
           Valsartan              20 to 40 mg twice                           160 mg twice
        254 mg/d

        The use of ARBs to reduce morbidity and mortality is recommended in patients with current
        or prior symptoms of chronic HFrEF who are intolerant to ACE inhibitors because of cough or
        angioedema (Class I, Level of Evidence A) (ACC/AHA/HFSA, 2017).

        ARBs are reasonable to reduce morbidity and mortality as alternatives to ACE inhibitors as
        first-line therapy for patients with HFrEF, especially for patients already taking ARBs for
        other indications, unless contraindicated (Class IIa, Level of Evidence: A) (ACCF/AHA,
        2013).

        Addition of an ARB may be considered in persistently symptomatic patients with HFrEF who
        are already being treated with an ACE inhibitor and a beta blocker in whom an aldosterone
        antagonist is not indicated or tolerated (Class IIb, Level of Evidence: A) (ACCF/AHA,
        2013).

        The clinical strategy of inhibition of the renin-angiotensin system with ACE inhibitors
        (Level of Evidence A), or ARBs (Level of Evidence A) or ARNI (Level of Evidence B-R) in
        conjunction with evidence-based beta-blockers, and aldosterone antagonists in selected
        patients, is recommended for patients with chronic HFrEF to reduce morbidity and mortality
        (Class I) (ACC/AHA/HFSA, 2017).

        In patients with chronic symptomatic HFrEF NYHA class II or III who tolerate an ACE
        inhibitor or ARB, replacement by an ARNI is recommended to further reduce morbidity and
        mortality (Class I, Level of Evidence: B-R) (ACC/AHA/HFSA, 2017).

        ARNI should not be administered concomitantly with ACE inhibitors or within 36 hours of the
        last dose of an ACE inhibitor (Class III, Level of Evidence B-R) (ACC/AHA/HFSA, 2017).

        ARNI should not be administered to patients with a history of angioedema (Class III
        Recommendation, Level of Evidence C-EO) (ACC/AHA/HFSA, 2017).

        For the hospitalized patient:
        In patients with HFrEF experiencing a symptomatic exacerbation of HF requiring
        hospitalization during chronic maintenance treatment with GDMT [guideline-directed medical
        therapy; GDMT represents optimal medical therapy as defined by ACCF/AHA guideline-
        recommended therapies (primarily Class I)], it is recommended that GDMT be continued in the
        absence of hemodynamic instability or contraindications (Class I, Level of Evidence: B)
        (ACCF/AHA, 2013).
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms134v6_diabetes_medical_attention_for_nephropathy import (
    ClinicalQualityMeasure134v6
)
from canvas_workflow_kit.patient_recordset import (
    BillingLineItemRecordSet,
    InstructionRecordSet,
    MedicationRecordSet,
    ProtocolOverrideRecordSet,
    ReferralReportRecordSet
)
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class ClinicalQualityMeasure134v6Test(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure134v6.enabled())

    def test_description(self):
        expected = (
            'Patients 18-75 years of age with diabetes who have not had a nephropathy screening '
            'test in the last year or evidence of nephropathy.')
        self.assertEqual(expected, ClinicalQualityMeasure134v6._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS134v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure134v6._meta.information)

    def test_version(self):
        self.assertTrue(hasattr(ClinicalQualityMeasure134v6._meta, 'version'))

    def test_change_types(self):
        result = ClinicalQualityMeasure134v6._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'instruction',
            'lab_report',
            'medication',
            'referral_report',
            'patient',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_in_numerator_with_override(self):
        patient = self.load_patient('cms134v6_diabetes_yes')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure134v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.lab_reports[0]['originalDate'] = '2018-08-10'

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=0))
        self.assertTrue(tested.in_numerator())
        expected = r'^Dohav has diabetes and had a Dialysis Related Service [\w\s]+ on 7/26/18'
        self.assertRegex(tested.message, expected)
        self.assertEqual(-1, tested._due_in)

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=80))
        self.assertTrue(tested.in_numerator())
        expected = r'^Dohav has diabetes and had a Dialysis Related Service [\w\s]+ on 7/26/18'
        self.assertRegex(tested.message, expected)
        self.assertEqual(-1, tested._due_in)

        del (patient.conditions.records[1])
        patient.referral_reports = ReferralReportRecordSet([])
        patient.medications = MedicationRecordSet([])
        patient.instructions = InstructionRecordSet([])

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=0))
        self.assertTrue(tested.in_numerator())
        expected = r'^Dohav has diabetes and a urine protein test was done [\w\s]+ on 8/10/18'
        self.assertRegex(tested.message, expected)
        self.assertEqual(175, tested._due_in)

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=80))
        self.assertTrue(tested.in_numerator())
        self.assertEqual(95, tested._due_in)

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=175))
        self.assertTrue(tested.in_numerator())
        self.assertEqual(0, tested._due_in)

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=176))
        self.assertFalse(tested.in_numerator())
        self.assertEqual(-1, tested._due_in)

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('cms134v6_diabetes_no')
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has diabetes --> true
        patient = self.load_patient('cms134v6_diabetes_yes')
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to young < 18 --> false
        patient = self.load_patient('cms134v6_diabetes_yes')
        patient.patient['birthDate'] = '2000-08-24'
        self.assertTrue(17 < patient.age_at(timeframe.end) < 18)
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '2000-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 18)
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to old > 75 --> false
        patient = self.load_patient('cms134v6_diabetes_yes')
        patient.patient['birthDate'] = '1943-08-22'
        self.assertTrue(75 < patient.age_at(timeframe.end) < 76)
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 75)
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-24'
        self.assertTrue(patient.age_at(timeframe.end) < 75)
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient has an other code --> false
        patient = self.load_patient('cms134v6_diabetes_yes')
        for condition in patient.conditions:
            for code in condition['coding']:
                code['code'] = 'code'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_numerator(self):
        now = arrow.get('2018-12-06 13:24:56')
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # not test when patient has no diabetes because it makes no sense

        # referral report
        patient = self.load_patient('cms134v6_diabetes_yes')
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = r'^Dohav has diabetes and had a Dialysis Related Service [\w\s]+ on 7/26/18'
        self.assertRegex(tested.message, expected)
        self.assertEqual(-1, tested._due_in)

        # medication
        patient.referral_reports = ReferralReportRecordSet([])
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = 'Dohav has diabetes and is under Ace Inhibitors medication'
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        # conditions - HypertensiveChronicKidneyDisease
        patient.medications = MedicationRecordSet([])

        patient.conditions.records[1]['coding'][0]['code'] = 'I129'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = 'Dohav has diabetes and has been diagnosed Hypertensive Chronic Kidney Disease'
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        patient.conditions.records[1]['coding'][0]['code'] = 'N170'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = 'Dohav has diabetes and has been diagnosed Kidney Failure'
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        patient.conditions.records[1]['coding'][0]['code'] = 'N000'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = ('Dohav has diabetes and has been diagnosed Glomerulonephritis '
                    'and Nephrotic Syndrome')
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        patient.conditions.records[1]['coding'][0]['code'] = 'E0829'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = 'Dohav has diabetes and has been diagnosed Diabetic Nephropathy'
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        patient.conditions.records[1]['coding'][0]['code'] = 'R808'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = 'Dohav has diabetes and has been diagnosed Proteinuria'
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        patient.conditions.records[1]['coding'][0]['system'] = 'http://snomed.info/sct'
        patient.conditions.records[1]['coding'][0]['code'] = '175901007'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = 'Dohav has diabetes and had a Kidney Transplant'
        self.assertEqual(expected, tested.message)
        self.assertEqual(-1, tested._due_in)

        # instructions
        del (patient.conditions.records[1])
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = (r'^Dohav has diabetes and had an ESRD Monthly Outpatient Services [\w\s]+ '
                    'on 7/6/18')
        self.assertRegex(tested.message, expected)
        self.assertEqual(-1, tested._due_in)

        # lab report
        patient.instructions = InstructionRecordSet([])
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        self.assertTrue(tested.in_numerator())
        expected = r'^Dohav has diabetes and a urine protein test was done [\w\s]+ on 8/5/18'
        self.assertRegex(tested.message, expected)
        self.assertEqual(241, tested._due_in)

    def test_compute_results(self):
        start = arrow.get('2017-10-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> n/a
        patient = self.load_patient('cms134v6_diabetes_no')
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

        # patient has diabetes --> satisfied
        patient = self.load_patient('cms134v6_diabetes_yes')
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = r'^Dohav has diabetes and had a Dialysis Related Service [\w\s]+ on 7/26/18$'
        self.assertRegex(result.narrative, text)
        self.assertEqual([], result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has diabetes --> satisfied
        patient = self.load_patient('cms134v6_diabetes_yes')

        del (patient.conditions.records[1])
        patient.referral_reports = ReferralReportRecordSet([])
        patient.medications = MedicationRecordSet([])
        patient.instructions = InstructionRecordSet([])

        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = r'^Dohav has diabetes and a urine protein test was done [\w\s]+ on 8/5/18$'
        self.assertRegex(result.narrative, text)
        self.assertEqual([], result.recommendations)
        self.assertEqual(285, result.due_in)  # not one year since the time frame define the period

        # patient has diabetes --> due
        patient = self.load_patient('cms134v6_diabetes_yes_only')
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (r'^Dohav has diabetes and a urine microalbumin test is due to screen '
                r'for nephropathy$')
        self.assertRegex(result.narrative, text)
        expected = [{
            'title': 'Order a urine microalbumin test',
            'command': {
                'type': 'labOrder',
                'filter': {
                    'coding': [
                        {
                            'system': 'loinc',
                            'code': [
                                '50561-0', '14956-7', '11218-5', '5804-0',
                                '57735-3', '44292-1', '53525-2', '53531-0',
                                '58992-9', '21482-5', '13801-6', '63474-1',
                                '9318-7', '12842-1', '43607-1', '20454-5',
                                '32209-9', '1755-8', '77158-4', '13705-9',
                                '20621-9', '2887-8', '56553-1', '53532-8',
                                '32551-4', '30003-8', '43606-3', '50949-7',
                                '14959-1', '53530-2', '1754-1', '40663-7',
                                '14585-4', '14957-5', '32294-1', '2889-4',
                                '49023-5', '57369-1', '2888-6', '43605-5',
                                '35663-4', '26801-1', '53121-0', '27298-9',
                                '30000-4', '40486-3', '40662-9', '2890-2',
                                '30001-2', '1757-4', '18373-1', '59159-4',
                                '1753-3', '77254-1', '58448-2', '60678-0',
                                '76401-9', '14958-3', '47558-2', '21059-1',
                                '34366-5', '77253-3',
                            ]
                        }
                    ]
                }
            },
            'key': 'CMS134v6_RECOMMEND_URINE_TEST',
            'rank': 1,
            'button': 'Order',

        }]  # yapf: disable
        # self.assertEqual(recommendations, result.recommendations)
        self.helper_compare_recommendations(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has diabetes + younger than 18 --> n/a + due in
        patient = self.load_patient('cms134v6_diabetes_no')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)
        patient = self.load_patient('cms134v6_diabetes_yes_only')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(6, result.due_in)

    def test_compute_results_with_override(self):
        patient = self.load_patient('cms134v6_diabetes_yes')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure134v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.lab_reports[0]['originalDate'] = '2018-08-10'

        # exam done within the last 180 days
        del (patient.conditions.records[1])
        patient.referral_reports = ReferralReportRecordSet([])
        patient.medications = MedicationRecordSet([])
        patient.instructions = InstructionRecordSet([])

        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = 'Dohav has diabetes and a urine protein test was done 2 months ago on 8/10/18'
        self.assertEqual(text, result.narrative)
        self.assertEqual(0, len(result.recommendations))
        self.assertEqual(105, result.due_in)

        # exam done 181 days
        patient.billing_line_items = BillingLineItemRecordSet([{
            'id': 92,
            'created': arrow.get('2018-08-01').shift(days=186),
            'datetimeOfService': arrow.get('2018-08-01').shift(days=186),
            'cpt': '99202',
            'units': 1
        }])
        tested = ClinicalQualityMeasure134v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=186))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = 'Dohav has diabetes and a urine microalbumin test is due to screen for nephropathy'
        self.assertEqual(text, result.narrative)
        self.assertEqual(1, len(result.recommendations))
        self.assertEqual(-1, result.due_in)

    def helper_compare_recommendations(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])
            exp_item = exp['command']['filter']['coding']
            rec_item = recommendation.command['filter']['coding']
            self.assertEqual(len(exp_item), len(rec_item))
            for exp_coding, rec_coding in zip(exp_item, rec_item):
                self.assertEqual(exp_coding['system'], rec_coding['system'])
                self.assertEqual(sorted(exp_coding['code']), sorted(rec_coding['code']))



# type: ignore
from typing import Dict, List

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events
from canvas_workflow_kit.canvas_code_set import CanvasCodeSet
from canvas_workflow_kit.patient_recordset import InterviewRecordSet
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.specials import (
    Covid19QuestionnaireHighRiskOutreach,
    Covid19QuestionnaireSymptomaticSurveillance
)


# flake8: noqa
class Ccp001v1(ClinicalQualityMeasure):

    class Meta:

        title = 'COVID-19 Risk Assessment Follow Up'
        version = "2020-04-03v1"
        changelog = "Initial release"

        description = 'All patients with COVID Questionnaire completed Date < 7 days ago and >  5 days ago.'
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol'

        identifiers = ['CCP001v1']

        types = ['CCP']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

        references = [
            'Canvas Medical CCP, https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol'
        ]

        show_in_chart = False

    anchor_format = 'YYYY-MM-DD 12:00:00ZZ'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.rationales_for_inclusion_in_numerator: List = []

    @cached_property
    def anchor_time(self) -> arrow:
        # regardless of the time of the day at the practice
        # the anchor time is set at the middle of day at the practice
        # all questionnaires from 12 of D-1 to 12 D will be considered on day D
        timezone = 'UTC'
        if 'timezone' in self.patient.patient:
            timezone = self.patient.patient['timezone']
        return arrow.get(self.timeframe.end.to(timezone).format(self.anchor_format)).to('utc')

    @cached_property
    def interview(self) -> Dict:
        start = self.anchor_time.shift(days=-6)
        period = Timeframe(start=start, end=self.anchor_time)
        return (self.patient.interviews.find(Covid19QuestionnaireSymptomaticSurveillance |
                                             Covid19QuestionnaireHighRiskOutreach).within(period).
                first() or {})

    def in_initial_population(self) -> bool:
        """
        Patients with COVID Questionnaire completed Date < 7 days ago
        """
        return bool(self.interview)

    def in_denominator(self) -> bool:
        """
        Patients with COVID Questionnaire completed 6 days ago
        """
        if not self.in_initial_population():
            return False

        start = self.anchor_time.shift(days=-6)
        end = start.shift(days=1)
        interview_date = arrow.get(self.interview[InterviewRecordSet.DATE_FIELD])
        return bool(start <= interview_date < end)

    def in_numerator(self) -> bool:
        """
        Patients that need a follow up
        """
        follow_up_question = CanvasCodeSet.code('CANVAS0005')
        follow_up_response = '183616001'
        symptom_question = CanvasCodeSet.code('CANVAS0002')
        travel_question = CanvasCodeSet.code('CANVAS0003')
        travel_response = '276030007'
        exposure_question = CanvasCodeSet.code('CANVAS0004')
        exposure_response = '150781000119103'

        result = False
        interview = self.interview
        if interview:
            question_responses = self.group_question_responses(interview)
            answers = {r['code']: r['value'] for r in interview['responses']}

            for question, responses in question_responses.items():
                rationale = ''
                if question == follow_up_question and follow_up_response in responses:
                    rationale = 'Follow-up requested by care team'
                elif question == symptom_question:
                    rationale = f'Symptoms exhibited: {", ".join([answers[c] for c in responses])}'

                if rationale:
                    result = True
                    self.rationales_for_inclusion_in_numerator.append(rationale)

            # Recent travel or recent exposure can only be a primary inclusion criteria if the patient is 65+.
            # If younger, it will be included as context along with other primary inclusion criteria.
            follow_up = result
            for question, responses in question_responses.items():
                rationale = ''
                if question == travel_question and travel_response in responses:
                    if follow_up:
                        rationale = 'Recent travel'
                    elif self.patient.age_at(self.anchor_time) >= 65:
                        rationale = 'Age 65+ with recent travel'
                elif question == exposure_question and exposure_response in responses:
                    if follow_up:
                        rationale = 'Recent exposure'
                    elif self.patient.age_at(self.anchor_time) >= 65:
                        rationale = 'Age 65+ with recent exposure'

                if rationale:
                    result = True
                    self.rationales_for_inclusion_in_numerator.append(rationale)

        return result

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            result.due_in = -1
            result.next_review = arrow.utcnow().shift(days=1)
            if self.in_numerator():
                result.status = STATUS_DUE
                result.add_narrative(f'{self.patient.first_name} should have a follow-up arranged')
                for rationale in self.rationales_for_inclusion_in_numerator:
                    result.add_narrative(rationale)

                result.add_recommendation(
                    Recommendation(
                        key='CCP001v1_RECOMMEND_FOLLOW_UP',
                        rank=1,
                        button='Communication',
                        title='Arrange a follow-up',
                        narrative=f'{self.patient.first_name} should have a follow-up arranged',
                        command={'key': 'schedule'}))
            else:
                result.status = STATUS_SATISFIED
                result.add_narrative(
                    f'{self.patient.first_name} does not need a follow-up arranged')
        elif self.in_initial_population():
            result.due_in = 5 - (
                arrow.utcnow() - arrow.get(self.interview[InterviewRecordSet.DATE_FIELD])).days

            now_local = arrow.utcnow().to(self.patient.patient['timezone'])
            tomorrow_noon = arrow.get(now_local.format(self.anchor_format))
            sometime_after_noon = tomorrow_noon.shift(days=1, hours=1)
            result.next_review = sometime_after_noon.to('utc')

        return result
